# Global import
# Global imports
import json

from easyeda2kicad.easyeda.parameters_easyeda import *


class easyeda_symbol_importer:
    def __init__(self, easyeda_cp_cad_data):
        self.input = easyeda_cp_cad_data
        self.output: ee_symbol = self.extract_easyeda_data(
            ee_data=easyeda_cp_cad_data,
            ee_data_info=easyeda_cp_cad_data["dataStr"]["head"]["c_para"],
        )

    def get_symbol(self):
        return self.output

    def extract_easyeda_data(self, ee_data: dict, ee_data_info: dict):
        new_ee_symbol = ee_symbol(
            info=ee_symbol_info(
                name=ee_data_info["name"],
                prefix=ee_data_info["pre"],
                package=ee_data_info.get("package", None),
                manufacturer=ee_data_info.get("BOM_Manufacturer", None),
                datasheet=ee_data["lcsc"].get("url", None),
                lcsc_id=ee_data["lcsc"].get("number", None),
                jlc_id=ee_data_info.get("BOM_JLCPCB Part Class", None),
            ),
            bbox=ee_symbol_bbox(
                x=float(ee_data["dataStr"]["head"]["x"]),
                y=float(ee_data["dataStr"]["head"]["y"]),
            ),
        )

        for line in ee_data["dataStr"]["shape"]:
            designator = line.split("~")[0]
            # For pins
            if designator == "P":
                ee_pin = self.extract_easyeda_pin(pin_data=line)
                new_ee_symbol.pins.append(self.tune_ee_pin(pin=ee_pin))
            # For rectangles
            elif designator == "R":
                ee_rectangle = self.extract_easyeda_rectangle(rectangle_data=line)
                new_ee_symbol.rectangles.append(
                    self.tune_ee_rectangle(rect=ee_rectangle)
                )
            # For polylines
            elif designator == "PL":
                ee_polyline = self.extract_easyeda_polyline(polyline_data=line)
                new_ee_symbol.polylines.append(
                    self.tune_ee_polyline(polyline=ee_polyline)
                )
            # For polygons
            elif designator == "PG":
                ee_polygon = self.extract_easyeda_polygon(polygon_data=line)
                new_ee_symbol.polygons.append(self.tune_ee_polygon(polygon=ee_polygon))
            # For paths
            elif designator == "PT":
                ee_path = self.extract_easyeda_path(path_data=line)
                new_ee_symbol.paths.append(self.tune_ee_path(path=ee_path))
            # For Pie
            elif designator == "PI":
                # Elliptical arc seems to be not supported in Kicad
                ...
            # For ellipse
            elif designator == "E":
                ...  # Ellipse seems to be not supported in Kicad
            # For arcs
            elif designator == "A":
                ...  # TODO
            else:
                print(f"\t[-] Unknow symbol designator : {designator}")

        return new_ee_symbol

    # ---------------------------------------

    def extract_easyeda_pin(self, pin_data: str):
        segments = pin_data.split("^^")
        ee_segments = [seg.split("~") for seg in segments]

        pin_settings = ee_symbol_pin_settings(
            **dict(zip(ee_symbol_pin_settings.__fields__, ee_segments[0][1:]))
        )
        pin_dot = ee_symbol_pin_dot(
            dot_x=float(ee_segments[1][0]), dot_y=float(ee_segments[1][1])
        )
        pin_path = ee_symbol_pin_path(path=ee_segments[2][0], color=ee_segments[2][1])
        pin_name = ee_symbol_pin_name(
            **dict(zip(ee_symbol_pin_name.__fields__, ee_segments[3][:]))
        )

        pin_dot_bis = ee_symbol_pin_dot_bis(
            is_displayed=ee_segments[5][0],
            circle_x=float(ee_segments[5][1]),
            circle_y=float(ee_segments[5][2]),
        )
        pin_clock = ee_symbol_pin_clock(
            is_displayed=ee_segments[6][0], path=ee_segments[6][1]
        )
        return ee_symbol_pin(
            settings=pin_settings,
            pin_dot=pin_dot,
            pin_path=pin_path,
            name=pin_name,
            dot=pin_dot_bis,
            clock=pin_clock,
        )

    # ---------------------------------------

    def extract_easyeda_rectangle(self, rectangle_data: str):
        ee_segment = rectangle_data.split("~")
        return ee_symbol_rectangle(
            **dict(zip(ee_symbol_rectangle.__fields__, ee_segment[1:]))
        )

    def extract_easyeda_polyline(self, polyline_data: str):
        ee_segment = polyline_data.split("~")
        return ee_symbol_polyline(
            **dict(zip(ee_symbol_polyline.__fields__, ee_segment[1:]))
        )

    def extract_easyeda_polygon(self, polygon_data: str):
        ee_segment = polygon_data.split("~")
        return ee_symbol_polygon(
            **dict(zip(ee_symbol_polygon.__fields__, ee_segment[1:]))
        )

    def extract_easyeda_path(self, path_data: str):
        ee_segment = path_data.split("~")
        return ee_symbol_path(**dict(zip(ee_symbol_path.__fields__, ee_segment[1:])))

    # ---------------------------------------

    def tune_ee_pin(self, pin: ee_symbol_pin):
        pin.settings.rotation = (
            pin.settings.rotation if pin.settings.rotation != "" else "0"
        )
        pin.settings.type = (
            easyeda_pin_type(int(pin.settings.type)).name
            if pin.settings.type != ""
            else "unspecified"
        )
        pin.pin_path.path = pin.pin_path.path.replace("v", "h")
        pin.name.text = pin.name.text.replace(" ", "")
        return pin

    def tune_ee_rectangle(self, rect: ee_symbol_rectangle):
        return rect

    def tune_ee_polyline(self, polyline: ee_symbol_polyline):
        return polyline

    def tune_ee_polygon(self, polygon: ee_symbol_polygon):
        return polygon

    def tune_ee_path(self, path: ee_symbol_path):
        return path

    # ---------------------------------------


# ------------------------------------------------------------------------------


class easyeda_footprint_importer:
    def __init__(self, easyeda_cp_cad_data):
        self.input = easyeda_cp_cad_data
        self.output = self.extract_easyeda_data(
            ee_data_str=self.input["packageDetail"]["dataStr"],
            ee_data_info=self.input["packageDetail"]["dataStr"]["head"]["c_para"],
        )

    def get_footprint(self):
        return self.output

    def extract_easyeda_data(self, ee_data_str: str, ee_data_info: str):
        new_ee_footprint = ee_footprint(
            info=ee_footprint_info(
                name=ee_data_info["package"],
                fp_type="smd" if "SMT" in ee_data_info else "tht",
            ),
            bbox=ee_footprint_bbox(
                x=float(ee_data_str["head"]["x"]),
                y=float(ee_data_str["head"]["y"]),
            ),
        )

        for line in ee_data_str["shape"]:

            ee_designator = line.split("~")[0]
            ee_fields = line.split("~")[1:]

            if ee_designator == "PAD":
                ee_pad = ee_footprint_pad(
                    **dict(zip(ee_footprint_pad.__fields__, ee_fields[:18]))
                )
                new_ee_footprint.pads.append(ee_pad)
            elif ee_designator == "TRACK":
                ee_track = ee_footprint_track(
                    **dict(zip(ee_footprint_track.__fields__, ee_fields))
                )
                new_ee_footprint.tracks.append(ee_track)
            elif ee_designator == "HOLE":
                ee_hole = ee_footprint_hole(
                    **dict(zip(ee_footprint_hole.__fields__, ee_fields))
                )
                new_ee_footprint.holes.append(ee_hole)
            elif ee_designator == "CIRCLE":
                ee_circle = ee_footprint_circle(
                    **dict(zip(ee_footprint_circle.__fields__, ee_fields))
                )
                new_ee_footprint.circles.append(ee_circle)
            elif ee_designator == "ARC":
                ee_arc = ee_footprint_arc(
                    **dict(zip(ee_footprint_arc.__fields__, ee_fields))
                )
                new_ee_footprint.arcs.append(ee_arc)
            elif ee_designator == "RECT":
                ee_rectangle = ee_footprint_rectangle(
                    **dict(zip(ee_footprint_rectangle.__fields__, ee_fields))
                )
                new_ee_footprint.rectangles.append(ee_rectangle)
            elif ee_designator == "TEXT":
                ee_text = ee_footprint_text(
                    **dict(zip(ee_footprint_text.__fields__, ee_fields))
                )
                new_ee_footprint.texts.append(ee_text)
            elif ee_designator == "SOLIDREGION":
                ...
            elif ee_designator == "SVGNODE":
                ...
            else:
                print(f"\t[-] Unknow footprint designator : {ee_designator}")

        return new_ee_footprint



# Global imports
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Union

from pydantic import BaseModel, validator


class easyeda_pin_type(Enum):
    unspecified = 0
    input = 1
    output = 2
    bidirectional = 3
    power = 4


# ------------------------- Symbol -------------------------


class ee_symbol_bbox(BaseModel):
    x: float
    y: float


# ---------------- PIN ----------------
class ee_symbol_pin_settings(BaseModel):
    is_displayed: bool
    type: str
    spice_pin_number: str
    pos_x: float
    pos_y: float
    rotation: int
    id: str
    is_locked: bool

    @validator("is_displayed", pre=True)
    def parse_display_field(cls, v):
        return True if v == "show" else v

    @validator("is_locked", pre=True)
    def empty_str_lock(cls, is_locked):
        return False if is_locked == "" else is_locked

    @validator("rotation", pre=True)
    def empty_str_rotation(cls, rotation):
        return 0.0 if rotation == "" else rotation


class ee_symbol_pin_dot(BaseModel):
    dot_x: float
    dot_y: float


@dataclass
class ee_symbol_pin_path:
    path: str
    color: str


class ee_symbol_pin_name(BaseModel):
    is_displayed: bool
    pos_x: float
    pos_y: float
    rotation: int
    text: str
    text_anchor: str
    font: str
    font_size: float

    @validator("font_size", pre=True)
    def empty_str_font(cls, font_size):
        if isinstance(font_size, str) and "pt" in font_size:
            return float(font_size.replace("pt", ""))
        return 7.0 if font_size == "" else font_size

    @validator("is_displayed", pre=True)
    def parse_display_field(cls, v):
        return True if v == "show" else v

    @validator("rotation", pre=True)
    def empty_str_rotation(cls, rotation):
        return 0.0 if rotation == "" else rotation


class ee_symbol_pin_dot_bis(BaseModel):
    is_displayed: bool
    circle_x: float
    circle_y: float

    @validator("is_displayed", pre=True)
    def parse_display_field(cls, v):
        return True if v == "show" else v


class ee_symbol_pin_clock(BaseModel):
    is_displayed: bool
    path: str

    @validator("is_displayed", pre=True)
    def parse_display_field(cls, v):
        return True if v == "show" else v


@dataclass
class ee_symbol_pin:
    settings: ee_symbol_pin_settings
    pin_dot: ee_symbol_pin_dot
    pin_path: ee_symbol_pin_path
    name: ee_symbol_pin_name
    dot: ee_symbol_pin_dot_bis
    clock: ee_symbol_pin_clock


# ---------------- RECTANGLE ----------------
class ee_symbol_rectangle(BaseModel):
    pos_x: float
    pos_y: float
    rx: Union[float, None]
    ry: Union[float, None]
    width: float
    height: float
    stroke_color: str
    stroke_width: str
    stroke_style: str
    fill_color: str
    id: str
    is_locked: bool

    @validator("*", pre=True)
    def empty_str_to_none(cls, v):
        return None if v == "" else v


# ---------------- POLYLINE ----------------
class ee_symbol_polyline(BaseModel):
    points: str
    stroke_color: str
    stroke_width: str
    stroke_style: str
    fill_color: str
    id: str
    is_locked: bool

    @validator("is_locked", pre=True)
    def empty_str_lock(cls, is_locked):
        return False if is_locked == "" else is_locked


# ---------------- POLYGON ----------------
class ee_symbol_polygon(ee_symbol_polyline):
    ...


# ---------------- PATH ----------------
# TODO : ee_symbol_path.paths should be a SVG PATH https://www.w3.org/TR/SVG11/paths.html#PathElement
# TODO : small svg parser and then convert to kicad
# TODO: support bezier curve, currently paths are seen as polygone
class ee_symbol_path(BaseModel):
    paths: str
    stroke_color: str
    stroke_width: str
    stroke_style: str
    fill_color: str
    id: str
    is_locked: bool

    @validator("is_locked", pre=True)
    def empty_str_lock(cls, is_locked):
        return False if is_locked == "" else is_locked

    # @validator("paths", pre=True)
    # def clean_svg_path(cls, paths:str):
    #     return paths.replace("M", "").replace("C","")


# ---------------- SYMBOL ----------------
@dataclass
class ee_symbol_info:
    name: str = ""
    prefix: str = ""
    package: str = ""
    manufacturer: str = ""
    datasheet: str = ""
    lcsc_id: str = ""
    jlc_id: str = ""


@dataclass
class ee_symbol:
    info: ee_symbol_info
    bbox: ee_symbol_bbox
    pins: List[ee_symbol_pin] = field(default_factory=list)
    rectangles: List[ee_symbol_rectangle] = field(default_factory=list)
    polylines: List[ee_symbol_polyline] = field(default_factory=list)
    polygons: List[ee_symbol_polygon] = field(default_factory=list)
    paths: List[ee_symbol_path] = field(default_factory=list)


# ------------------------- Footprint -------------------------


def convert_to_mm(dim: float):
    return float(dim) * 10 * 0.0254


@dataclass
class ee_footprint_bbox:

    x: float
    y: float

    def convert_to_mm(self):
        self.x = convert_to_mm(self.x)
        self.y = convert_to_mm(self.y)


class ee_footprint_pad(BaseModel):
    shape: str
    center_x: float
    center_y: float
    width: float
    height: float
    layer_id: int
    net: str
    number: str
    hole_radius: float
    points: str
    rotation: float
    id: str
    hole_length: float
    hole_point: str
    is_plated: bool
    is_locked: bool

    def convert_to_mm(self):
        self.center_x = convert_to_mm(self.center_x)
        self.center_y = convert_to_mm(self.center_y)
        self.width = convert_to_mm(self.width)
        self.height = convert_to_mm(self.height)
        self.hole_radius = convert_to_mm(self.hole_radius)
        self.hole_length = convert_to_mm(self.hole_length)

    @validator("is_locked", pre=True)
    def empty_str_lock(cls, is_locked):
        return False if is_locked == "" else is_locked

    @validator("rotation", pre=True)
    def empty_str_rotation(cls, rotation):
        return 0.0 if rotation == "" else rotation


class ee_footprint_track(BaseModel):
    stroke_width: float
    layer_id: int
    net: str
    points: str
    id: str
    is_locked: bool

    @validator("is_locked", pre=True)
    def empty_str_lock(cls, is_locked):
        return False if is_locked == "" else is_locked

    def convert_to_mm(self):
        self.stroke_width = convert_to_mm(self.stroke_width)


class ee_footprint_hole(BaseModel):
    center_x: float
    center_y: float
    radius: float
    id: str
    is_locked: bool

    @validator("is_locked", pre=True)
    def empty_str_lock(cls, is_locked):
        return False if is_locked == "" else is_locked

    def convert_to_mm(self):
        self.center_x = convert_to_mm(self.center_x)
        self.center_y = convert_to_mm(self.center_y)
        self.radius = convert_to_mm(self.radius)


class ee_footprint_circle(BaseModel):
    cx: float
    cy: float
    radius: float
    stroke_width: float
    layer_id: int
    id: str
    is_locked: bool

    @validator("is_locked", pre=True)
    def empty_str_lock(cls, is_locked):
        return False if is_locked == "" else is_locked

    def convert_to_mm(self):
        self.cx = convert_to_mm(self.cx)
        self.cy = convert_to_mm(self.cy)
        self.radius = convert_to_mm(self.radius)
        self.stroke_width = convert_to_mm(self.stroke_width)


class ee_footprint_rectangle(BaseModel):
    x: float
    y: float
    width: float
    height: float
    stroke_width: float
    id: str
    layer_id: int
    is_locked: bool

    @validator("is_locked", pre=True)
    def empty_str_lock(cls, is_locked):
        return False if is_locked == "" else is_locked

    def convert_to_mm(self):
        self.x = convert_to_mm(self.x)
        self.y = convert_to_mm(self.y)
        self.width = convert_to_mm(self.width)
        self.height = convert_to_mm(self.height)


class ee_footprint_arc(BaseModel):
    stroke_width: float
    layer_id: int
    net: str
    path: str
    helper_dots: str
    id: str
    is_locked: bool

    @validator("is_locked", pre=True)
    def empty_str_lock(cls, is_locked):
        return False if is_locked == "" else is_locked


class ee_footprint_text(BaseModel):
    type: str
    center_x: float
    center_y: float
    stroke_width: float
    rotation: int
    miror: str
    layer_id: int
    net: str
    font_size: float
    text: str
    text_path: str
    is_displayed: bool
    id: str
    is_locked: bool

    @validator("is_displayed", pre=True)
    def empty_str_display(cls, is_displayed):
        return True if is_displayed == "" else is_displayed

    @validator("is_locked", pre=True)
    def empty_str_lock(cls, is_locked):
        return False if is_locked == "" else is_locked

    @validator("rotation", pre=True)
    def empty_str_rotation(cls, rotation):
        return 0.0 if rotation == "" else rotation

    def convert_to_mm(self):

        self.center_x = convert_to_mm(self.center_x)
        self.center_y = convert_to_mm(self.center_y)
        self.stroke_width = convert_to_mm(self.stroke_width)
        self.font_size = convert_to_mm(self.font_size)


# ---------------- FOOTPRINT ----------------


@dataclass
class ee_footprint_info:
    name: str
    fp_type: str


@dataclass
class ee_footprint:
    info: ee_footprint_info
    bbox: ee_footprint_bbox
    pads: List[ee_footprint_pad] = field(default_factory=list)
    tracks: List[ee_footprint_track] = field(default_factory=list)
    holes: List[ee_footprint_hole] = field(default_factory=list)
    circles: List[ee_footprint_circle] = field(default_factory=list)
    arcs: List[ee_footprint_arc] = field(default_factory=list)
    rectangles: List[ee_footprint_rectangle] = field(default_factory=list)
    texts: List[ee_footprint_text] = field(default_factory=list)



# Global imports
import itertools
from dataclasses import dataclass, field, fields
from enum import Enum
from typing import List

# ---------------- CONFIG ----------------

# Origin point.
KI_XO = 0
KI_YO = 0

# Settings for box drawn around pins in a unit.
KI_DEFAULT_BOX_LINE_WIDTH = 0

# Mapping from understandable schematic symbol box fill-type name
# to the fill-type indicator used in the KiCad part library.
KI_BOX_FILLS = {"no_fill": "N", "fg_fill": "F", "bg_fill": "f"}
KI_DEFAULT_BOX_FILL = "bg_fill"

# Part reference.
KI_REF_SIZE = 60  # Font size.
KI_REF_Y_OFFSET = 200

# Part number.
KI_PART_NUM_SIZE = 60  # Font size.
KI_PART_NUM_Y_OFFSET = 200

# Part footprint
KI_PART_FOOTPRINT_SIZE = 60  # Font size.
KI_PART_FOOTPRINT_Y_OFFSET = KI_PART_NUM_Y_OFFSET + 100

# Part manufacturer number.
KI_PART_MPN_SIZE = 60  # Font size.
KI_PART_MPN_Y_OFFSET = KI_PART_FOOTPRINT_Y_OFFSET + 100

# Part datasheet.
KI_PART_DATASHEET_SIZE = 60  # Font size.
KI_PART_DATASHEET_Y_OFFSET = KI_PART_MPN_Y_OFFSET + 100

# Part description.
KI_PART_DESC_SIZE = 60  # Font size.
KI_PART_DESC_Y_OFFSET = KI_PART_DATASHEET_Y_OFFSET + 100


KI_ROTATION = {"left": 0, "right": 180, "bottom": 90, "top": -90}

KI_PIN_LENGTH = 100
KI_PIN_SPACING = 100
KI_PIN_NUM_SIZE = 50  # Font size for pin numbers.
KI_PIN_NAME_SIZE = 50  # Font size for pin names.
KI_PIN_NAME_OFFSET = 40  # Separation between pin and pin name.
KI_PIN_ORIENTATION = "left"
KI_PIN_STYLE = "line"
KI_SHOW_PIN_NUMBER = True  # Show pin numbers when True.
KI_SHOW_PIN_NAME = True  # Show pin names when True.
KI_SINGLE_PIN_SUFFIX = ""
KI_MULTI_PIN_SUFFIX = "*"
KI_PIN_SPACER_PREFIX = "*"

# --------------

KI_LIB_HEADER = "EESchema-LIBRARY Version 2.3\n#encoding utf-8\n"
KI_START_DEF = "DEF {name} {ref} 0 {pin_name_offset} {show_pin_number} {show_pin_name} {num_units} L N\n"
KI_REF_FIELD = 'F0 "{ref_prefix}" {x} {y} {font_size} H V {text_justification} CNN\n'
KI_PARTNUM_FIELD = 'F1 "{num}" {x} {y} {font_size} H V {text_justification} CNN\n'
KI_FOOTPRINT_FIELD = (
    'F2 "{footprint}" {x} {y} {font_size} H I {text_justification} CNN\n'
)
KI_DATASHEET_FIELD = (
    'F3 "{datasheet}" {x} {y} {font_size} H I {text_justification} CNN\n'
)
KI_MPN_FIELD = 'F4 "{manufacturer}" 0 0 0 H I C CNN "Manufacturer"\n'
KI_DESC_FIELD = 'F5 "{desc}" 0 0 0 H I C CNN "desc"\n'
KI_LCSC_FIELD = 'F6 "{id}" 0 0 0 H I C CNN "LCSC Part"\n'
KI_JLCPCB_FIELD = 'F7 "{type}" 0 0 0 H I C CNN "JLC Part"\n'

KI_START_DRAW = "DRAW\n"


KI_END_DRAW = "ENDDRAW\n"
KI_END_DEF = "ENDDEF\n"

# --------------

KI_PIN_TYPES = {
    "input": "I",
    "inp": "I",
    "in": "I",
    "clk": "I",
    "output": "O",
    "outp": "O",
    "out": "O",
    "bidirectional": "B",
    "bidir": "B",
    "bi": "B",
    "inout": "B",
    "io": "B",
    "iop": "B",
    "tristate": "T",
    "tri": "T",
    "passive": "P",
    "pass": "P",
    "unspecified": "U",
    "un": "U",
    "": "U",
    "analog": "U",
    "power_in": "W",
    "pwr_in": "W",
    "pwrin": "W",
    "power": "W",
    "pwr": "W",
    "ground": "W",
    "gnd": "W",
    "power_out": "w",
    "pwr_out": "w",
    "pwrout": "w",
    "pwr_o": "w",
    "open_collector": "C",
    "opencollector": "C",
    "open_coll": "C",
    "opencoll": "C",
    "oc": "C",
    "open_emitter": "E",
    "openemitter": "E",
    "open_emit": "E",
    "openemit": "E",
    "oe": "E",
    "no_connect": "N",
    "noconnect": "N",
    "no_conn": "N",
    "noconn": "N",
    "nc": "N",
}

KI_PIN_STYLES = {
    "line": "",
    "inverted": "I",
    "clock": "C",
    "input_low": "L",
    "inp_low": "L",
    "in_lw": "L",
    "in_b": "L",
    "in_n": "L",
    "~in": "L",
    "#in": "L",
    "clock_low": "CL",
    "clk_low": "CL",
    "clk_lw": "CL",
    "output_low": "V",
    "outp_low": "V",
    "out_lw": "V",
    "out_b": "V",
    "out_n": "V",
    "~out": "V",
    "#out": "V",
    "falling_edge_clock": "F",
    "falling_clk": "F",
    "fall_clk": "F",
    "non_logic": "X",
    "nl": "X",
    "analog": "X",
}

KI_PIN_ORIENTATIONS = {
    "": "R",
    "left": "R",
    "right": "L",
    "bottom": "U",
    "down": "U",
    "top": "D",
    "up": "D",
}


class kicad_pin_orientation(Enum):
    right = 0
    top = 90
    left = 180
    bottom = 270


# ---------------------------- SYMBOL PART ----------------------------
# ---------------- INFO HEADER ----------------
@dataclass
class ki_symbol_info:
    name: str
    prefix: str
    package: str
    manufacturer: str
    datasheet: str
    lcsc_id: str
    jlc_id: str
    y_low: int = 0
    y_high: int = 0

    def export(self) -> str:
        return "\n".join(
            (
                f"#\n# {self.name}\n#",
                "DEF {name} {ref} 0 {pin_name_offset} {show_pin_number} {show_pin_name} {num_units} L N".format(
                    name=self.name,
                    ref=self.prefix,
                    pin_name_offset=KI_PIN_NAME_OFFSET,
                    show_pin_number=KI_SHOW_PIN_NUMBER and "Y" or "N",
                    show_pin_name=KI_SHOW_PIN_NAME and "Y" or "N",
                    num_units=1,
                ),
                'F0 "{ref_prefix}" {x} {y} {font_size} H V {text_justification} CNN'.format(
                    ref_prefix=self.prefix,
                    x=0,
                    y=self.y_high + KI_REF_Y_OFFSET,
                    text_justification="C",  # Center align
                    font_size=KI_REF_SIZE,
                ),
                'F1 "{num}" {x} {y} {font_size} H V {text_justification} CNN'.format(
                    num=self.name,
                    x=0,
                    y=self.y_low - KI_PART_NUM_Y_OFFSET,
                    text_justification="C",  # Center align
                    font_size=KI_PART_NUM_SIZE,
                ),
                'F2 "{footprint}" {x} {y} {font_size} H I {text_justification} CNN'.format(
                    footprint=self.package,
                    x=0,
                    y=self.y_low - KI_PART_FOOTPRINT_Y_OFFSET,
                    text_justification="C",  # Center align
                    font_size=KI_PART_FOOTPRINT_SIZE,
                )
                if self.package
                else "",
                'F3 "{datasheet}" {x} {y} {font_size} H I {text_justification} CNN'.format(
                    datasheet=self.datasheet,
                    x=0,
                    y=self.y_low - KI_PART_DATASHEET_Y_OFFSET,
                    text_justification="C",  # Center align
                    font_size=KI_PART_DATASHEET_SIZE,
                )
                if self.datasheet
                else "",
                'F4 "{manufacturer}" 0 0 0 H I C CNN "Manufacturer"'.format(
                    manufacturer=self.manufacturer,
                )
                if self.manufacturer
                else "",
                f'F6 "{self.lcsc_id}" 0 0 0 H I C CNN "LCSC Part"'
                if self.lcsc_id
                else "",
                f'F7 "{self.jlc_id}" 0 0 0 H I C CNN "JLC Part"' if self.jlc_id else "",
                "DRAW\n",
            )
        ).replace("\n\n", "\n")


# ---------------- PIN ----------------
@dataclass
class ki_symbol_pin:
    name: str
    number: str
    style: str
    type: float
    orientation: float
    pos_x: int
    pos_y: int

    def export(self) -> str:
        return "X {name} {num} {x} {y} {length} {orientation} {num_sz} {name_sz} {unit_num} 1 {pin_type} {pin_style}\n".format(
            name=self.name,
            num=self.number,
            x=self.pos_x,
            y=self.pos_y,
            length=KI_PIN_LENGTH,
            orientation=self.orientation,
            num_sz=KI_PIN_NUM_SIZE,
            name_sz=KI_PIN_NAME_SIZE,
            unit_num=1,
            pin_type=self.type,
            pin_style=self.style,
        )


# ---------------- RECTANGLE ----------------
@dataclass
class ki_symbol_rectangle:
    pos_x0: int = 0
    pos_y0: int = 0
    pos_x1: int = 0
    pos_y1: int = 0

    def export(self) -> str:
        return "S {x0} {y0} {x1} {y1} {unit_num} 1 {line_width} {fill}\n".format(
            x0=int(round(self.pos_x0 / 50.0)) * 50,
            y0=int(round(self.pos_y0 / 50.0)) * 50,
            x1=int(round(self.pos_x1 / 50.0)) * 50,
            y1=int(round(self.pos_y1 / 50.0)) * 50,
            unit_num=1,
            line_width=KI_DEFAULT_BOX_LINE_WIDTH,
            fill=KI_BOX_FILLS["bg_fill"],
        )


# ---------------- POLYGON ----------------
@dataclass
class ki_symbol_polygon:
    points: List[List[int]] = field(default_factory=List[List[int]])
    points_number: int = 0
    is_closed: bool = False

    def export(self) -> str:
        return (
            "P {points_number} {unit_num} 1 {line_width} {coordinate} {fill}\n".format(
                points_number=self.points_number,
                unit_num=1,
                line_width=KI_DEFAULT_BOX_LINE_WIDTH,
                coordinate=" ".join(list(itertools.chain.from_iterable(self.points))),
                fill=KI_BOX_FILLS["bg_fill"]
                if self.is_closed
                else KI_BOX_FILLS["no_fill"],
            )
        )


# ---------------- CIRCLE ----------------
@dataclass
class ki_symbol_circle:
    pos_x: int = 0
    pos_y: int = 0
    radius: int = 0

    def export(self) -> str:
        return "C {pos_x} {pos_y} {radius} {unit_num} 1 {line_width} {fill}\n".format(
            pos_x=self.pos_x,
            pos_y=self.pos_y,
            radius=self.radius,
            unit_num=1,
            line_width=KI_DEFAULT_BOX_LINE_WIDTH,
            fill=KI_BOX_FILLS["bg_fill"],
        )


# ---------------- ARC ----------------
@dataclass
class ki_symbol_arc:
    pos_x: int = 0
    pos_y: int = 0
    radius: int = 0
    angle_start: float = 0.0
    angle_end: float = 0.0
    start_x: int = 0
    start_y: int = 0
    end_x: int = 0
    end_y: int = 0

    def export(self) -> str:
        return "C {pos_x} {pos_y} {radius} {angle_start} {angle_end} {unit_num} 1 {line_width} {fill} {start_x} {start_y} {end_x} {end_y}\n".format(
            pos_x=self.pos_x,
            pos_y=self.pos_y,
            radius=self.radius,
            angle_start=self.angle_start,
            angle_end=self.angle_end,
            unit_num=1,
            line_width=KI_DEFAULT_BOX_LINE_WIDTH,
            fill=KI_BOX_FILLS["bg_fill"]
            if self.angle_start == self.angle_end
            else KI_BOX_FILLS["no_fill"],
            start_x=self.start_x,
            start_y=self.start_y,
            end_x=self.end_x,
            end_y=self.end_y,
        )


# ---------------- BEZIER CURVE ----------------
@dataclass
class ki_symbol_bezier:
    points: List[List[int]] = field(default_factory=List[List[int]])
    points_number: int = 0
    is_closed: bool = False

    def export(self) -> str:
        return (
            "B {points_number} {unit_num} 1 {line_width} {coordinate} {fill}\n".format(
                points_number=self.points_number,
                unit_num=1,
                line_width=KI_DEFAULT_BOX_LINE_WIDTH,
                coordinate=" ".join(list(itertools.chain.from_iterable(self.points))),
                fill=KI_BOX_FILLS["bg_fill"]
                if self.is_closed
                else KI_BOX_FILLS["no_fill"],
            )
        )


# ---------------- SYMBOL ----------------
@dataclass
class ki_symbol:
    info: ki_symbol_info
    pins: List[ki_symbol_pin] = field(default_factory=lambda: [])
    rectangles: List[ki_symbol_rectangle] = field(default_factory=lambda: [])
    circles: List[ki_symbol_circle] = field(default_factory=lambda: [])
    arcs: List[ki_symbol_arc] = field(default_factory=lambda: [])
    polygons: List[ki_symbol_polygon] = field(default_factory=lambda: [])
    beziers: List[ki_symbol_bezier] = field(default_factory=lambda: [])

    def export(self) -> str:
        lib_output = ""
        # Get y_min and y_max to put component info
        self.info.y_low = min(pin.pos_y for pin in self.pins) if self.pins else 0
        self.info.y_high = max(pin.pos_y for pin in self.pins) if self.pins else 0

        for _field in fields(self):
            shapes = getattr(self, _field.name)
            if isinstance(shapes, list):
                for sub_symbol in shapes:
                    lib_output += sub_symbol.export()
            else:
                lib_output += shapes.export()

        lib_output += "ENDDRAW\n" + "ENDDEF\n"

        return lib_output


# ---------------------------- FOOTPRINT PART ----------------------------

KI_MODULE_INFO = "(module {package_lib}:{package_name} (layer F.Cu) (tedit {edit})\n"
KI_DESCRIPTION = (
    '\t(descr "{datasheet_link}, generated with easyeda2kicad.py on {date}")\n'
)
KI_TAGS_INFO = '\t(tags "{tag}")\n'
KI_FP_TYPE = "\t(attr {component_type})\n"
KI_REFERENCE = "\t(fp_text reference REF** (at {pos_x} {pos_y}) (layer F.SilkS)\n\t\t(effects (font (size 1 1) (thickness 0.15)))\n\t)\n"
KI_PACKAGE_VALUE = "\t(fp_text value {package_name} (at {pos_x} {pos_y}) (layer F.Fab)\n\t\t(effects (font (size 1 1) (thickness 0.15)))\n\t)\n"
KI_FAB_REF = "\t(fp_text user %R (at 0 0) (layer F.Fab)\n\t\t(effects (font (size 1 1) (thickness 0.15)))\n\t)\n"
KI_END_FILE = ")"

KI_PAD = "\t(pad {number} {type} {shape} (at {pos_x:.2f} {pos_y:.2f} {orientation:.2f}) (size {width:.2f} {height:.2f}) (layers {layers}){drill}{polygon})\n"
KI_LINE = "\t(fp_line (start {start_x:.2f} {start_y:.2f}) (end {end_x:.2f} {end_y:.2f}) (layer {layers}) (width {stroke_width:.2f}))\n"
KI_HOLE = '\t(pad "" thru_hole circle (at {pos_x:.2f} {pos_y:.2f}) (size {size:.2f} {size:.2f}) (drill {size:.2f}) (layers *.Cu *.Mask))\n'
KI_CIRCLE = "\t(fp_circle (center {cx:.2f} {cy:.2f}) (end {end_x:.2f} {end_y:.2f}) (layer {layers}) (width {stroke_width:.2f}))\n"
KI_ARC = "\t(fp_arc (start {start_x:.2f} {start_y:.2f}) (end {end_x:.2f} {end_y:.2f}) (angle {angle:.2f}) (layer {layers}) (width {stroke_width:.2f}))\n"
KI_TEXT = "\t(fp_text user {text} (at {pos_x:.2f} {pos_y:.2f} {orientation:.2f}) (layer {layers}){display}\n\t\t(effects (font (size {font_size:.2f} {font_size:.2f}) (thickness {thickness:.2f})) (justify left{mirror}))\n\t)\n"

# ---------------------------------------

KI_PAD_SHAPE = {
    "ELLIPSE": "circle",
    "RECT": "rect",
    "OVAL": "oval",
    "POLYGON": "custom",
}
KI_PAD_LAYER = {
    1: "F.Cu F.Paste F.Mask",
    2: "B.Cu B.Paste B.Mask",
    3: "F.SilkS",
    11: "*.Cu *.Paste *.Mask",
    13: "F.Fab",
    15: "Dwgs.User",
}

KI_LAYERS = {
    1: "F.Cu",
    2: "B.Cu",
    3: "F.SilkS",
    4: "B.SilkS",
    5: "F.Paste",
    6: "B.Paste",
    7: "F.Mask",
    8: "B.Mask",
    10: "Edge.Cuts",
    11: "Edge.Cuts",
    12: "Cmts.User",
    13: "F.Fab",
    14: "B.Fab",
    15: "Dwgs.User",
    101: "F.Fab",
}


# Round all float values contained in the dataclass
def round_float_values(self):
    for _field in fields(self):
        current_value = getattr(self, _field.name)
        if isinstance(current_value, float):
            setattr(self, _field.name, round(current_value, 2))


# ---------------- PAD ----------------
@dataclass
class ki_footprint_pad:
    type: str
    shape: str
    pos_x: float
    pos_y: float
    width: float
    height: float
    layers: str
    number: str
    drill: float
    orientation: float
    polygon: str

    def __post_init__(self):
        round_float_values(self)


# ---------------- TRACK ----------------
@dataclass
class ki_footprint_track:
    points_start_x: List[float] = field(default_factory=list)
    points_start_y: List[float] = field(default_factory=list)
    points_end_x: List[float] = field(default_factory=list)
    points_end_y: List[float] = field(default_factory=list)
    stroke_width: float = 0
    layers: str = ""


# ---------------- HOLE ----------------
@dataclass
class ki_footprint_hole:
    pos_x: float
    pos_y: float
    size: float

    def __post_init__(self):
        round_float_values(self)


# ---------------- CIRCLE ----------------
@dataclass
class ki_footprint_circle:
    cx: float
    cy: float
    end_x: float
    end_y: float
    layers: str
    stroke_width: float

    def __post_init__(self):
        round_float_values(self)


# ---------------- RECTANGLE ----------------
@dataclass
class ki_footprint_rectangle(ki_footprint_track):
    ...
    # points_start_x:List[float] = field(default_factory=list)
    # points_start_y:List[float] = field(default_factory=list)
    # points_end_x:List[float] = field(default_factory=list)
    # points_end_y:List[float] = field(default_factory=list)
    # stroke_width:float = 0
    # layers:str = ''


# ---------------- ARC ----------------
@dataclass
class ki_footprint_arc:
    start_x: float
    start_y: float
    end_x: float
    end_y: float
    angle: float
    layers: str
    stroke_width: float

    def __post_init__(self):
        round_float_values(self)


# ---------------- TEXT ----------------
@dataclass
class ki_footprint_text:
    pos_x: float
    pos_y: float
    orientation: float
    text: str
    layers: str
    font_size: float
    thickness: float
    display: str
    mirror: str

    def __post_init__(self):
        round_float_values(self)


# ---------------- VIA ----------------
@dataclass
class ki_footprint_via:
    name: str = ""
    # TODO


# ---------------- SOLID REGION ----------------
@dataclass
class ki_footprint_solid_region:
    name: str = ""
    # TODO


# ---------------- COPPER AREA ----------------
@dataclass
class ki_footprint_copper_area:
    name: str = ""
    # TODO


# ---------------- FOOTPRINT INFO ----------------
@dataclass
class ki_footprint_info:
    name: str
    fp_type: str


# ---------------- FOOTPRINT OBJ ----------------
@dataclass
class ki_footprint:
    info: ki_footprint_info
    pads: List[ki_footprint_pad] = field(default_factory=list)
    tracks: List[ki_footprint_track] = field(default_factory=list)
    vias: List[ki_footprint_via] = field(default_factory=list)
    holes: List[ki_footprint_hole] = field(default_factory=list)
    circles: List[ki_footprint_circle] = field(default_factory=list)
    arcs: List[ki_footprint_arc] = field(default_factory=list)
    rectangles: List[ki_footprint_rectangle] = field(default_factory=list)
    texts: List[ki_footprint_text] = field(default_factory=list)
    solid_regions: List[ki_footprint_solid_region] = field(default_factory=list)
    copper_areas: List[ki_footprint_copper_area] = field(default_factory=list)



# Global imports
import argparse
import os
import re
import sys
from typing import List

from easyeda2kicad.easyeda.easyeda_api import easyeda_api
from easyeda2kicad.easyeda.easyeda_importer import (
    easyeda_footprint_importer,
    easyeda_symbol_importer,
)
from easyeda2kicad.easyeda.parameters_easyeda import ee_symbol
from easyeda2kicad.kicad.export_kicad_footprint import exporter_footprint_kicad
from easyeda2kicad.kicad.export_kicad_symbol import exporter_symbol_kicad


def get_parser() -> argparse.ArgumentParser:

    parser = argparse.ArgumentParser(
        description="A Python script that convert any electronic components from LCSC or EasyEDA to a Kicad library"
    )

    parser.add_argument("--lcsc_id", help="LCSC id", required=True, type=str)

    parser.add_argument(
        "--symbol", help="Get symbol of this id", required=False, action="store_true"
    )

    parser.add_argument(
        "--footprint",
        help="Get footprint of this id",
        required=False,
        action="store_true",
    )
    parser.add_argument(
        "--3d", help="Get 3d model of this id", required=False, action="store_true"
    )

    parser.add_argument(
        "--output",
        required=False,
        metavar="file.lib",
        help="Output file",
        type=str,
        # default=
    )

    parser.add_argument(
        "--overwrite",
        required=False,
        help="overwrite symbol and footprint lib if there is already a component with this lcsc_id",
        action="store_true",
    )

    return parser


def valid_arguments(arguments: dict) -> bool:

    if not arguments["lcsc_id"].startswith("C"):
        print("[-] Error: lcsc_id should start by C....")
        return False

    if not any([arguments["symbol"], arguments["footprint"], arguments["3d"]]):
        print("[-] Error: Missing action arguments. For example :")
        print("  easyeda2kicad --lcsc_id=C2040 --footprint")
        print("  easyeda2kicad --lcsc_id=C2040 --symbol")
        return False

    if arguments["output"]:
        base_folder = "/".join(arguments["output"].replace("\\", "/").split("/")[:-1])
        lib_name = (
            arguments["output"].replace("\\", "/").split("/")[-1].split(".lib")[0]
        )
        # Check input
        if not os.path.isdir(base_folder):
            print("Can't find the folder")
            return False
    else:
        default_folder = os.path.join(
            os.path.expanduser("~"), "Documents", "Kicad", "easyeda2kicad"
        )
        if not os.path.isdir(default_folder):
            os.mkdir(default_folder)

        base_folder = default_folder
        lib_name = "easyeda2kicad"

    arguments["output"] = f"{base_folder}/{lib_name}"
    # Create new lib files if they dont exist
    if not os.path.isdir(f"{arguments['output']}.pretty"):
        os.mkdir(f"{arguments['output']}.pretty")
        print(f"[+] Create {lib_name}.pretty footprint folder in {base_folder}")

    if not os.path.isfile(f"{arguments['output']}.lib"):
        with open(
            file=f"{arguments['output']}.lib", mode="w+", encoding="utf-8"
        ) as my_lib:
            my_lib.write("EESchema-LIBRARY Version 2.4\n#encoding utf-8\n")
        print(f"[+] Create {lib_name}.lib symbol lib in {base_folder}")

    return True


def id_already_in_symbol_lib(
    lib_path: str, component_id: str, component_name: str
) -> bool:
    with open(lib_path, encoding="utf-8") as f:
        current_lib = f.read()
        component = re.findall(
            rf'(#\n# {component_name}\n#\n.*?F6 "{component_id}".*?ENDDEF)',
            current_lib,
            flags=re.DOTALL,
        )

        if component != []:
            print(f"[*] This id is already in {lib_path}")
            return True
    return False


def delete_component_in_symbol_lib(
    lib_path: str, component_id: str, component_name: str
):
    with open(file=lib_path, encoding="utf-8") as f:
        current_lib = f.read()
        new_data = re.sub(
            rf'(#\n# {component_name}\n#\n.*?F6 "{component_id}".*?ENDDEF\n)',
            "",
            current_lib,
            flags=re.DOTALL,
        )

    with open(file=lib_path, mode="w", encoding="utf-8") as my_lib:
        my_lib.write(new_data)


def fp_already_in_footprint_lib(lib_path: str, package_name: str) -> bool:
    if os.path.isfile(f"{lib_path}/{package_name}.kicad_mod"):
        print(f"[*] The footprint for this id is already in {lib_path}")
        return True
    return False


def main(argv: List[str] = sys.argv[1:]) -> int:

    # cli interface
    parser = get_parser()
    try:
        args = parser.parse_args(argv)
    except SystemExit as err:
        return err.code
    arguments = vars(args)

    if not valid_arguments(arguments=arguments):
        return 1

    component_id = arguments["lcsc_id"]
    print("-- easyeda2kicad.py --")

    # Get CAD data of the component using easyeda API
    api = easyeda_api()
    cad_data = api.get_cad_data_of_component(lcsc_id=component_id)

    #   ---------------- SYMBOL ----------------
    if arguments["symbol"]:
        importer = easyeda_symbol_importer(easyeda_cp_cad_data=cad_data)
        easyeda_symbol: ee_symbol = importer.get_symbol()

        is_id_already_in_symbol_lib = id_already_in_symbol_lib(
            lib_path=f"{arguments['output']}.lib",
            component_id=easyeda_symbol.info.lcsc_id,
            component_name=easyeda_symbol.info.name,
        )
        if not arguments["overwrite"] and is_id_already_in_symbol_lib:
            print("[-] Error: Use --overwrite to update the older symbol lib")
            return 1

        print(f"[*] Creating Kicad symbol library for LCSC id : {component_id}")

        exporter = exporter_symbol_kicad(symbol=easyeda_symbol, kicad_version=5)
        # print(exporter.output)
        kicad_symbol_lib = exporter.get_kicad_lib()

        if is_id_already_in_symbol_lib:
            delete_component_in_symbol_lib(
                lib_path=f"{arguments['output']}.lib",
                component_id=easyeda_symbol.info.lcsc_id,
                component_name=easyeda_symbol.info.name,
            )

        with open(
            file=f"{arguments['output']}.lib", mode="a+", encoding="utf-8"
        ) as my_lib:
            my_lib.write(kicad_symbol_lib)

    # ---------------- FOOTPRINT ----------------
    if arguments["footprint"]:
        importer = easyeda_footprint_importer(easyeda_cp_cad_data=cad_data)
        easyeda_footprint = importer.get_footprint()

        is_id_already_in_footprint_lib = fp_already_in_footprint_lib(
            lib_path=f"{arguments['output']}.pretty",
            package_name=easyeda_footprint.info.name,
        )
        if not arguments["overwrite"] and is_id_already_in_footprint_lib:
            print("[-] Error: Use --overwrite to replace the older footprint lib")
            return 1

        print(f"[*] Creating Kicad footprint library for LCSC id : {component_id}")
        exporter = exporter_footprint_kicad(footprint=easyeda_footprint)
        kicad_footprint_lib, lib_name = exporter.export_footprint()

        # print(kicad_footprint_lib)
        with open(
            file=f"{arguments['output']}.pretty/{lib_name}.kicad_mod",
            mode="w",
            encoding="utf-8",
        ) as my_lib:
            my_lib.write(kicad_footprint_lib)

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))
